from uplogic.animation import ActionSystem
from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode


class ULStopAction(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_layer = None
        self.act_system = self.get_act_system()
        self.OUT = self.add_output(self.get_done)
        
    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ActionSystem('default')

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)

        action = self.act_system.get_layer(game_object, action_layer)
        if action is not None:
            self.act_system.remove(action)
        self._done = True



from mathutils import Vector
from mathutils import Euler
from mathutils import Matrix
from bge import logic
from bge.types import KX_GameObject as GameObject
import math


def matmul (a, b):
    return a @ b


def multadd (a, b, c):
    return a * b + c


def isqrt (a):
    return a ** 1/2


def smin(a, b, c):
    if (c != 0):
        h = max(c - abs(a - b), 0) / c
        return min(a, b) - h * h * h * c * (.166666666666666)
    else:
        return min(a, b)


def smax(a, b, c):
    return -smin(-a, -b, c)


def sign(a):
    return (
        a if a == 0 else (
            1.0 if a > 0 else -1.0
        )
    )


def compare(a, b, c):
    return abs(a-b) <= c


def fraction(a):
    return a - math.floor(a)


def trunc_mod(a, b):
    return math.trunc(a % b)


def floor_mod(a, b):
    return math.floor(a % b)


def wrap(value, max, min):
    _range = max - min
    return value - (_range * math.floor((value - min) / _range)) if _range != 0 else min


def snap(value, step):
    if (step == 0):
        return  0.0
    else:
        return math.floor(value / step) * step


def ping_pong(value, scale):
    if (scale == 0.0):
        return 0.0
    return abs(fraction((value - scale) / (scale * 2.0)) * scale * 2.0 - scale)


def _min(a, b):
    return min(a, b)


def _max(a, b):
    return max(a, b)


def _round(value):
    return round(value)


def _log(value, base):
    return math.log(value, base)


def _acos(value):
    return math.acos(clamp(value))


def _asin(value):
    return math.asin(clamp(value))


def _atan(value):
    return math.atan(clamp(value))

def _lerp(a, b, fac):
    return lerp(a, b, fac)


def clamp(value: float, lower: float = 0, upper: float = 1) -> float:
    """Clamp a value in between two other values.

    :param `value`: input value
    :param `min`: minimum value
    :param `max`: maximum value

    :returns: clamped value as float
    """
    if isinstance(value, Vector):
        return vec_clamp(value, min, max)
    return max(lower, min(value, upper))


def cycle(value: float, min: float = 0, max: float = 1) -> float:
    """Clamp a value in between two other values.

    :param `value`: input value
    :param `min`: minimum value
    :param `max`: maximum value

    :returns: clamped value as float
    """
    if isinstance(value, Vector):
        return vec_clamp(value, min, max)
    if value >= max:
        return value - max
    if value < min:
        return max - abs(value - min)
    return value


def vec_clamp(vec: Vector, min: float = 0, max: float = 1) -> Vector:
    """Clamp length of a vector.

    :param `value`: `Vector`
    :param `min`: minimum length
    :param `max`: maximum length

    :returns: clamped vector as float
    """
    vec = vec.copy()
    if vec.length < min:
        vec.normalize()
        return vec * min
    if vec.length > max:
        vec.normalize()
        return vec * max
    return vec


def interpolate(a: float, b: float, fac: float, threshold: float = 0.001) -> float:
    """Interpolate between 2 values using a factor.

    :param `a`: starting value
    :param `b`: target value
    :param `fac`: interpolation factor

    :returns: calculated value as float
    """
    if -threshold < a-b < threshold:
        return b
    return (fac * b) + ((1-fac) * a)


def lerp(a: float, b: float, fac: float, threshold: float = 0.001) -> float:
    """Interpolate between 2 values using a factor.

    :param `a`: starting value
    :param `b`: target value
    :param `fac`: interpolation factor

    :returns: calculated value as float
    """
    if -threshold < a-b < threshold:
        return b
    return (fac * b) + ((1-fac) * a)


def vec_abs(vec):
    """Make every value of this vector positive.\n
    Only supports less than 4 Dimensions.

    :param `a`: `Vector`

    :returns: positive vector
    """
    vec = vec.copy()
    vec.x = abs(vec.x)
    vec.y = abs(vec.y)
    vec.z = abs(vec.z)
    return vec


def get_angle(a: Vector, b: Vector, up=Vector((0, 0, 1))) -> float:
    """Get the angle between the direction from a to b and up.

    :param `a`: `Vector` a
    :param `b`: `Vector` b
    :param `up`: compare direction

    :returns: calculated value as float
    """
    direction = get_direction(Vector(a), Vector(b))
    rad: float = direction.angle(up)
    deg: float = rad * 180/math.pi
    return deg


def get_raw_angle(a: Vector, b: Vector) -> float:
    """Get the angle between the direction from a to b and up.

    :param `a`: `Vector` a
    :param `b`: `Vector` b
    :param `up`: compare direction

    :returns: calculated value as float
    """
    rad: float = a.angle(b)
    deg: float = rad * 180/math.pi
    return deg


def angle_signed(a: Vector, b: Vector, up: Vector) -> float:
    return math.atan2(a.cross(b).dot(up), a.dot(b))

def get_direction(a, b, local=False) -> Vector:
    """Get the direction from one vector to another.

    :param `a`: `Vector` a
    :param `b`: `Vector` b
    :param `local`: use local space (position only)

    :returns: direction as `Vector`
    """
    start = a.worldPosition.copy() if hasattr(a, "worldPosition") else a
    if hasattr(b, "worldPosition"):
        b = b.worldPosition.copy()
    if local:
        b = start + b
    d = b - start
    d.normalize()
    return d


def map_range(value: float, in_min: float, in_max: float, out_min: float, out_max: float, clamp: bool = False) -> float:
    """Map a value from one range to another.
    
    :param `value`: Value to be remapped.
    :param `in_min`: Lower end of the original range.
    :param `in_max`: Upper end of the original range.
    :param `out_min`: Lower end of the new range.
    :param `out_max`: Upper end of the new range.
    :param `clamp`: Clamp the modified value.
    """
    result = (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
    if out_min > out_max:
        out_min, out_max = out_max, out_min
    if clamp and result < out_min:
        return out_min
    if clamp and result > out_max:
        return out_max
    return result


def map_range_vector(value: float, in_min: float, in_max: float, out_min: float, out_max: float, clamp: bool = False) -> float:
    """Map a vector from one range to another.

    :param `value`: Value to be remapped.
    :param `in_min`: Lower end of the original range.
    :param `in_max`: Upper end of the original range.
    :param `out_min`: Lower end of the new range.
    :param `out_max`: Upper end of the new range.
    :param `clamp`: Clamp the modified value.
    """
    outvec = Vector(value)
    for i in range(len(value)):
        result = (value[i] - in_min[i]) * (out_max[i] - out_min[i]) / (in_max[i] - in_min[i]) + out_min[i]
        if clamp and result < out_min[i]:
            result = out_min[i]
        if clamp and result > out_max[i]:
            result = out_max[i]
        outvec[i] = result
    return outvec


def get_local(obj, target) -> Vector:
    return obj.worldTransform.inverted() @ target


def get_bitmask(
    *slots: int, all=False
) -> int:
    """Get the collision bitmask value for the provided slot indices. Slots range from 0 to 15.
    
    :param `slots`: Arbitrary arguments, slots from 0-15 as int.
    :param `all`: Get the bitmask value of all slots combined, ignores `slots` argument."""
    if not all and not slots:
        return 0
    mask = 0
    for slot in range(16) if all else slots:
        mask += 1 << slot
    return mask

def get_collision_bitmask(
    *slots: int, all=False
) -> int:
    return get_bitmask(*slots, all)


def project_vector3(v, xi, yi):
    return Vector((v[xi], v[yi]))


def rotate2d(origin, pivot, angle):
    angle = math.radians(angle)
    return Vector((
        ((origin[0] - pivot[0]) * math.cos(angle)) - ((origin[1] - pivot[1]) * math.sin(angle)) + pivot[0],
        ((origin[0] - pivot[0]) * math.sin(angle)) + ((origin[1] - pivot[1]) * math.cos(angle)) + pivot[1]
    ))


def rotate3d(origin, pivot, angle, axis=2):
    angle = math.radians(angle)
    if axis == 0:
        return Vector((
            origin[0],
            ((origin[1] - pivot[1]) * math.cos(angle)) - ((origin[2] - pivot[2]) * math.sin(angle)) + pivot[1],
            ((origin[1] - pivot[1]) * math.sin(angle)) + ((origin[2] - pivot[2]) * math.cos(angle)) + pivot[2]
        ))
    elif axis == 1:
        return Vector((
            ((origin[0] - pivot[0]) * math.cos(angle)) - ((origin[2] - pivot[2]) * math.sin(angle)) + pivot[0],
            origin[1],
            ((origin[0] - pivot[0]) * math.sin(angle)) + ((origin[2] - pivot[2]) * math.cos(angle)) + pivot[2]
        ))
    elif axis == 2:
        return Vector((
            ((origin[0] - pivot[0]) * math.cos(angle)) - ((origin[1] - pivot[1]) * math.sin(angle)) + pivot[0],
            ((origin[0] - pivot[0]) * math.sin(angle)) + ((origin[1] - pivot[1]) * math.cos(angle)) + pivot[1],
            origin[2]
        ))
    return origin


def rotate_by_axis(origin: Vector, pivot: Vector, angle: float, axis: Vector):

    angle = math.radians(angle)

    origin = origin.copy() - pivot

    z_null = Vector((axis.x, axis.y))
    if z_null.length:
        thz = z_null.angle_signed(Vector((1, 0)))
    else:
        thz = 90

    axis = Vector((
            (axis[0] * math.cos(-thz)) - (axis[1] * math.sin(-thz)),
            (axis[0] * math.sin(-thz)) + (axis[1] * math.cos(-thz)),
            axis[2]
    ))
    target_point = Vector((
            (origin[0] * math.cos(-thz)) - (origin[1] * math.sin(-thz)),
            (origin[0] * math.sin(-thz)) + (origin[1] * math.cos(-thz)),
            origin[2]
    ))

    thy = axis.angle(Vector((0, 0, 1)))

    target_point = Vector((
            (target_point[0] * math.cos(thy)) - (target_point[2] * math.sin(thy)),
            target_point[1],
            (target_point[0] * math.sin(thy)) + (target_point[2] * math.cos(thy))
    ))

    target_point = Vector((
            (target_point[0] * math.cos(angle)) - (target_point[1] * math.sin(angle)),
            (target_point[0] * math.sin(angle)) + (target_point[1] * math.cos(angle)),
            target_point[2]
    ))

    target_point = Vector((
            (target_point[0] * math.cos(-thy)) - (target_point[2] * math.sin(-thy)),
            target_point[1],
            (target_point[0] * math.sin(-thy)) + (target_point[2] * math.cos(-thy))
    ))

    target_point = Vector((
            (target_point[0] * math.cos(thz)) - (target_point[1] * math.sin(thz)),
            (target_point[0] * math.sin(thz)) + (target_point[1] * math.cos(thz)),
            target_point[2]
    ))

    return pivot + target_point


def rotate_by_euler(origin: Vector, pivot: Vector, angles: Euler):

    origin = origin.copy() - pivot
    angles = Euler(angles)

    transmat = angles.to_matrix()

    return pivot + origin @ transmat



from uplogic.nodes import ULConditionNode


class ULOr(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = False
        self.cb = False
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return (
            self.get_input(self.ca)
            or
            self.get_input(self.cb)
        )


class ULOrList(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = False
        self.cb = False
        self.cc = False
        self.cd = False
        self.ce = False
        self.cf = False
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        conds = [
            self.get_input(self.ca),
            self.get_input(self.cb),
            self.get_input(self.cc),
            self.get_input(self.cd),
            self.get_input(self.ce),
            self.get_input(self.cf)
        ]
        return True in conds



from uplogic.nodes import ULOutSocket, ULParameterNode
from uplogic.utils import make_valid_name
from bpy.types import Object


class ULGetLogicTreeProperty(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.property_name = None
        self.OUT = self.add_output(self.get_property)

    def get_property(self):
        property_name = self.get_input(self.property_name)
        result = getattr(self.network.component, make_valid_name(property_name).lower(), False)
        if isinstance(result, Object):
            result = self.network.scene.getGameObjectFromObject(result)
        return result



from uplogic.animation import ActionSystem
from uplogic.animation.action import Action
from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode
from bpy.types import Action as BPYAction
from bge.types import KX_GameObject


class ULSetActionFrame(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_layer = None
        self.action_frame = None
        self.action_name = None
        self.layer_weight = None
        self.freeze = None
        self.act_system = self.get_act_system()
        self.OUT = self.add_output(self.get_done)

    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ActionSystem('default')

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object: KX_GameObject = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        action_frame = self.get_input(self.action_frame)
        freeze = self.get_input(self.freeze)
        bpy_action: BPYAction = self.get_input(self.action_name)
        intensity = self.get_input(self.layer_weight)
        action: Action = self.act_system.get_layer(game_object, action_layer)
        same_action = action is not None and action.name == bpy_action.name
        if not same_action or action is None:
            start_frame = bpy_action.frame_range[0]
            end_frame = bpy_action.frame_range[1]
            action = Action(
                game_object,
                bpy_action.name,
                start_frame,
                end_frame,
                action_layer,
                intensity=intensity
            )

        game_object.setActionFrame(action_frame, action_layer)
        action._intensity = intensity
        if freeze:
            action._speed=.00000000001
        action._restart_action()
        self._done = True



from bge import logic
from bge import render
from mathutils import Vector
from uplogic.nodes import ULActionNode, results
from uplogic.utils import get_bitmask
from uplogic.utils.raycasting import raycast_projectile, RayCastProjectileData


# @results('ray_data')
class ULProjectileRayCast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.origin = None
        self.destination = None
        self.local: bool = None
        self.power: float = None
        self.resolution: float = None
        self.property_name: str = None
        self.xray: bool = None
        self.mask: int = get_bitmask(all=True)
        self.distance: float = None
        self.visualize: bool = None
        self.network = None
        self.ray_data = RayCastProjectileData((None, None, None, None))
        self.RESULT = self.add_output(self.get_result)
        self.PICKED_OBJECT = self.add_output(self.get_picked_object)
        self.POINT = self.add_output(self.get_point)
        self.NORMAL = self.add_output(self.get_normal)
        self.PARABOLA = self.add_output(self.get_parabola)

    def setup(self, network):
        self.network = network

    def get_result(self):
        return self.ray_data.obj is not None

    def get_picked_object(self):
        return self.ray_data.obj

    def get_parabola(self):
        return self.ray_data.points

    def get_point(self):
        return self.ray_data.point

    def get_normal(self):
        return self.ray_data.normal

    def calc_projectile(self, t, vel, pos):
        half: float = logic.getCurrentScene().gravity * (.5 * t * t)
        vel = vel * t
        return half + vel + pos

    def reset(self):
        super().reset()
        self.ray_data = RayCastProjectileData((None, None, None, None))

    def evaluate(self):
        if not self.get_condition():
            return
        origin = self.get_input(self.origin)
        power: float = self.get_input(self.power)
        destination = self.get_input(self.destination)
        resolution: float = 1 - (self.get_input(self.resolution) * .99)
        property_name: str = self.get_input(self.property_name)
        xray: bool = self.get_input(self.xray)
        distance: float = self.get_input(self.distance)
        visualize: bool = self.get_input(self.visualize)
        destination.normalize()

        destination *= power
        origin = getattr(origin, 'worldPosition', origin)

        self.ray_data = raycast_projectile(
            caster=logic.getCurrentScene().active_camera,
            origin=Vector(origin),
            aim=Vector(destination),
            local=self.get_input(self.local),
            power=power,
            distance=distance,
            resolution=resolution,
            prop=property_name,
            xray=xray,
            mask=self.get_input(self.mask),
            visualize=visualize
        )



from bge import logic
from bge import events
from bge.types import SCA_InputEvent


KEYBOARD_EVENTS = logic.keyboard.inputs
'''Reference to `bge.logic.keyboard.inputs`
'''

_keys_active = {}


def key_event(key: str) -> SCA_InputEvent:
    '''Retrieve key event.\n
    Not intended for manual use.
    '''
    if isinstance(key, int):
        key = KEYBOARD_EVENTS[key]
    else:
        key = KEYBOARD_EVENTS[
        getattr(
            events, f'{key.upper()}KEY',
            (getattr(events, f'PAD{key.upper()}', None))
        )
    ]
    if key:
        return key
    else:
        print(f"""
        Key {key} not in [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
        `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
        `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
        `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
        `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
        `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
        `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
        `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
        `'SLASH'`, `'SPACE'`, `'TAB'`]
        """)
        return 0


def pad_event(key: str) -> bool:
    '''Retrieve Numpad event.\n
    Not intended for manual use.
    '''
    key = KEYBOARD_EVENTS[getattr(events, f'PAD{key}')]
    if key:
        return key
    else:
        print(f"""
        Key {key} not in [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
        `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
        `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
        `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
        `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
        `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
        `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
        `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
        `'SLASH'`, `'SPACE'`, `'TAB'`]
        """)
        return 0


def key_tap(key: str) -> bool:
    '''Detect key tapped.

    :param `key`: key as `str` of
    [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
    `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
    `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
    `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
    `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
    `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
    `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
    `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
    `'SLASH'`, `'SPACE'`, `'TAB'`]

    :returns: boolean
    '''
    return key_event(key).activated


def key_down(key: str) -> bool:
    '''Detect key held down.

    :param `key`: key as `str` of
    [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
    `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
    `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
    `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
    `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
    `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
    `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
    `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
    `'SLASH'`, `'SPACE'`, `'TAB'`]

    :returns: boolean
    '''
    key = key_event(key)
    return key.active or key.activated


def key_press(key: str, down=False):
    '''Detect key tap or held down.

    :param `key`: key as `str` of
    [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
    `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
    `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
    `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
    `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
    `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
    `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
    `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
    `'SLASH'`, `'SPACE'`, `'TAB'`]

    :returns: boolean
    '''
    key = key_event(key)
    return key.active or key.activated if down else key.activated



def key_up(key: str) -> bool:
    '''Detect key released.

    :param `key`: key as `str` of
    [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
    `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
    `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
    `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
    `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
    `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
    `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
    `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
    `'SLASH'`, `'SPACE'`, `'TAB'`]

    :returns: boolean
    '''
    return key_event(key).released


def key_pulse(key: str, time: float = .4) -> bool:
    '''Detect key tapped, then held down after `time` has passed.

    :param `key`: key as `str` of
    [`'A'`, `'B'`, `'C'`, `'D'`, `'E'`, `'F'`, `'G'`, `'H'`, `'I'`, `'J'`, `'K'`, `'L'`, `'M'`, `'N'`, `'O'`, `'P'`, `'Q'`,
    `'R'`, `'S'`, `'T'`, `'U'`, `'V'`, `'W'`, `'X'`, `'Y'`, `'Z'`, `'ZERO'`, `'ONE'`, `'TWO'`, `'THREE'`, `'FOUR'`, `'FIVE'`,
    `'SIX'`, `'SEVEN'`, `'EIGHT'`, `'NINE'`, `'CAPSLOCK'`, `'LEFTCTRL'`, `'LEFTSHIFT'` `'LEFTARROW'`, `'DOWNARROW'`, `'RIGHTARROW'`,
    `'UPARROW'`, `'0'`, `'1'`, `'2'`, `'3'`, `'4'`, `'5'`, `'6'`, `'7'`, `'8'`, `'9'`, `'PADPERIOD'`, `'PADSLASH'`, `'PADASTER'`,
    `'PADMINUS'`, `'PADENTER'`, `'PADPLUS'`, `'F1'`, `'F2'`, `'F3'`, `'F4'`, `'F5'`, `'F6'`, `'F7'`, `'F8'`, `'F9'`, `'F10'`,
    `'F11'`, `'F12'`, `'F13'`, `'F14'`, `'F15'`, `'F16'`, `'F17'`, `'F18'`, `'F19'`, `'ACCENTGRAVE'`, `'BACKSLASH'`,
    `'BACKSPACE'`, `'COMMA'`, `'DEL'`, `'END'`, `'EQUAL'`, `'ESC'`, `'HOME'`, `'INSERT'`, `'LEFTBRACKET'`, `'RIGHTBRACKET'`,
    `'LINEFEED'`, `'MINUS'`, `'PAGEDOWN'`, `'PAGEUP'`, `'PAUSE'`, `'PERIOD'`, `'QUOTE'`, `'RET'`, `'ENTER'`, `'SEMICOLON'`,
    `'SLASH'`, `'SPACE'`, `'TAB'`]
    :param time: timeout for key down

    :returns: boolean
    '''
    evt = key_event(key)
    if evt.activated:
        _keys_active[key] = 0
        return True
    k = _keys_active.get(key, 0)
    _keys_active[key] = k + (1 / (logic.getAverageFrameRate() or 0.01))
    if _keys_active[key] > time:
        return evt.active
    return False


class RecordedCharacter(tuple):

    @property
    def pressed(self):
        return self[0]

    @property
    def keycode(self):
        return self[1]

    @property
    def character(self):
        return self[2]


def record_keyboard(down=True, all=False) -> list[RecordedCharacter[bool, int, str]]:
    '''Listen to keyboard events.

    :param `down`: Record if key is down.
    :param `all`: Record non-character keys.

    :returns: List of tuples of `(pressed, keycode, character)`'''
    left_shift = KEYBOARD_EVENTS[events.LEFTSHIFTKEY].active
    right_shift = KEYBOARD_EVENTS[events.RIGHTSHIFTKEY].active
    active_events = logic.keyboard.activeInputs.copy()

    func = key_down if down else key_pulse

    evts = []

    for keycode in active_events:
        if func(keycode):
            event = active_events[keycode]
            char = events.EventToCharacter(
                event.type,
                left_shift or right_shift
            )
            evts.append(RecordedCharacter((
                True if char or all else False,
                keycode,
                char
            )))
    return evts


def keyboard_active() -> bool:
    return len(KEYBOARD_EVENTS) > 0


from uplogic.nodes import ULActionNode
from uplogic.ui import RelativeLayout, FloatLayout, BoxLayout, GridLayout
from uplogic.ui import PolarLayout as _PolarLayout
from math import degrees


class PolarLayout(_PolarLayout):
    def __init__(
        self,
        pos: list = (0, 0),
        size=(0, 0),
        bg_color=(0, 0, 0, 0),
        border_width=0,
        border_color=(0, 0, 0, 0),
        halign='center',
        valign='center',
        relative: dict = {},
        starting_angle: str = 0,
        radius: int = 100,
        angle: float = 0
    ):
        super().__init__(pos, relative, starting_angle, radius, angle)

layouts = {
    'FloatLayout': FloatLayout,
    'RelativeLayout': RelativeLayout,
    'BoxLayout': BoxLayout,
    'GridLayout': GridLayout,
    'PolarLayout': PolarLayout
}


class ULCreateUILayout(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.parent = None
        self.rel_pos = False
        self.pos = (0, 0)
        self.rel_size = False
        self.size = (100, 100)
        self.angle = 0
        self.color = (0, 0, 0, 0)
        self.border_width = 0
        self.border_color = (0, 0, 0, 0)
        self.spacing = 0
        self.cols = 1
        self.rows = 1
        self.starting_angle = 0
        self.radius = 100
        self._widget = None
        self.layout_type = 'RelativeLayout'
        self.boxlayout_type = 'vertical'
        self.halign_type = 'left'
        self.valign_type = 'bottom'
        self.OUT = self.add_output(self._get_done)
        self.WIDGET = self.add_output(self._get_widget)

    def _get_done(self):
        return self._done

    def _get_widget(self):
        return self._widget

    def evaluate(self):
        if not self.get_condition():
            return
        ipt = self.get_input
        parent = ipt(self.parent)
        rel_pos = ipt(self.rel_pos)
        pos = ipt(self.pos)
        rel_size = ipt(self.rel_size)
        size = ipt(self.size)
        angle = degrees(ipt(self.angle))
        color = ipt(self.color)
        border_width = ipt(self.border_width)
        border_color = ipt(self.border_color)

        self._widget = layouts[self.layout_type](
            pos=pos,
            size=size,
            bg_color=color,
            relative={'pos': rel_pos, 'size': rel_size},
            border_width=border_width,
            border_color=border_color,
            halign=self.halign_type,
            valign=self.valign_type,
            angle=angle
        )
        if self.layout_type == 'BoxLayout':
            self._widget.orientation = self.boxlayout_type
            self._widget.spacing = ipt(self.spacing)
        elif self.layout_type == 'GridLayout':
            self._widget.orientation = self.boxlayout_type
            self._widget.spacing = ipt(self.spacing)
            self._widget.cols = ipt(self.cols)
            self._widget.rows = ipt(self.rows)
        elif self.layout_type == 'PolarLayout':
            self._widget.starting_angle = self.starting_angle
            self._widget.radius = ipt(self.radius)
        if parent:
            parent.add_widget(self._widget)
        self._done = True
        



from .widget import Widget
import blf
from bpy.types import VectorFont
from uplogic.utils.math import rotate2d
import math


class Label(Widget):
    '''Widget for displaying text

    :param `pos`: Initial position of this widget in either pixels or factor.
    :param `relative`: Whether to use pixels or factor for size or pos; example: `{'pos': True, 'size': True}`.
    :param `text`: Initial text for this label.
    :param `font`: Font name.
    :param `font_color`: Color in RGBA.
    :param `font_size`: Font size in pt or factor.
    :param `line_height`: Total line height relative to character size (1 is same as character height).
    :param `shadow`: Draw a shadow behind the text.
    :param `shadow_offset`: Relative position of the shadow in px.
    :param `shadow_color`: Shadow color in RGBA.
    :param `halign`: Horizontal alignment of the widget, can be (`left`, `center`, `right`).
    :param `valign`: Vertical alignment of the widget, can be (`bottom`, `center`, `top`).
    :param `wrap`: Split lines that are too long for the containing widget.
    :param `angle`: Rotation in degrees of this widget around the pivot defined by the alignment.
    '''

    def __init__(
        self,
        pos=[0, 0],
        relative={},
        text='',
        font='',
        font_color=[1., 1., 1., 1.],
        font_size=12,
        line_height=1.5,
        shadow=False,
        shadow_offset=[1, -1],
        shadow_color=[.0, .0, .0, 1.],
        halign='left',
        valign='bottom',
        wrap=False,
        angle=0
    ):
        self._parent = None
        self._children = None
        self._font_color = font_color
        self.text = text
        self.line_height = line_height
        self.shadow = shadow
        self.shadow_offset = shadow_offset
        self.shadow_color = shadow_color
        self.font_size = font_size
        self.font_color = font_color
        self.font = font
        self.wrap = wrap
        self.lines = []
        Widget.__init__(self, pos, (0, 0), (0, 0, 0, 0), relative, angle=angle)
        self.text_halign = halign
        self.text_valign = valign
        self.start()

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, val):
        self._text = str(val)

    # @property
    # def halign(self):
    #     return self.text_halign

    # @halign.setter
    # def halign(self, val):
    #     self.text_halign = str(val)

    # @property
    # def valign(self):
    #     return self.text_valign

    # @valign.setter
    # def valign(self, val):
    #     self.text_valign = str(val)

    @property
    def pos_abs(self):
        return self._draw_pos

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, val):
        if isinstance(val, VectorFont):
            val = val.filepath.replace('\\', '/')
        self._font = blf.load(val) if val else 0

    @property
    def font_color(self):
        return self._font_color

    @font_color.setter
    def font_color(self, val):
        val = list(val)
        self._font_color = val

    @property
    def color(self):
        return self.font_color

    @color.setter
    def color(self, val):
        self.font_color = val

    # @property
    # def opacity(self):
    #     return self.bg_color[3]

    # @opacity.setter
    # def opacity(self, val):
    #     self.bg_color[3] = val
    #     self.font_opacity = val

    # @property
    # def font_opacity(self):
    #     return self._font_color[3]

    # @font_opacity.setter
    # def font_opacity(self, val):
    #     self._font_color[3] = val * self.opacity

    @property
    def dimensions(self):
        dim = blf.dimensions(self.font, self.text)
        lines = len(self.lines) or 1
        return (dim[0], blf.dimensions(self.font, 'Aj')[1] * lines)

    @property
    def _draw_size(self):
        relative = self.relative.get('font_size', False)
        fontsize = self.parent._draw_size[1] * self.font_size if relative else self.font_size
        blf.size(self.font, fontsize)
        return self.dimensions

    def make_floating(self, pos=True, size=True, halign='center', valign='center'):
        self.relative['pos'] = pos
        self.relative['size'] = size
        self.text_halign = halign
        self.text_valign = valign
        return self

    def draw(self):
        super()._setup_draw()
        if self.parent is None:
            return
        parsize = self.parent._draw_size
        relative = self.relative.get('font_size', False)
        font = self.font
        blf.size(font, parsize[1] * self.font_size if relative else self.font_size)
        col = self.font_color
        blf.color(font, col[0], col[1], col[2], col[3] * self.opacity)
        charsize = blf.dimensions(font, 'A')
        smallsize = blf.dimensions(font, 'a')[1]
        lowsize = blf.dimensions(font, 'g')[1]
        diff = lowsize - smallsize

        if self.angle or self.parent._draw_angle:
            blf.enable(font, blf.ROTATION)
            blf.rotation(font, math.radians(self._draw_angle))
        if self.parent.use_clipping:
            verts = self.parent._vertices
            blf.enable(font, blf.CLIPPING)
            blf.clipping(font, verts[0][0], verts[0][1] + charsize[1], verts[2][0], verts[2][1])
        else:
            blf.disable(font, blf.CLIPPING)
        if self.wrap and self.parent:
            blf.enable(font, blf.WORD_WRAP)
            blf.word_wrap(font, int(parsize[0]))
        if self.shadow:
            col = self.shadow_color
            blf.enable(font, blf.SHADOW)
            blf.shadow(font, 0, col[0], col[1], col[2], col[3] * self.opacity)
            blf.shadow_offset(font, int(self.shadow_offset[0]), int(self.shadow_offset[1]))
        lines = [t for t in self.text.split('\n')]
        if len(lines) > 1:
            for i, txt in enumerate(lines):
                pos = self._draw_pos.copy()
                dimensions = blf.dimensions(font, txt)
                underground = dimensions[1] > charsize[1]
                lheight = (charsize[1] * self.line_height)
                if self.text_halign == 'center':
                    pos[0] -= (dimensions[0] * .5)
                elif self.text_halign == 'right':
                    pos[0] -= dimensions[0]
                if self.text_valign == 'top':
                    # if underground:
                    # pos[1] += (diff)
                    pos[1] -= lheight
                elif self.text_valign == 'center':
                    # if underground:
                        # pos[1] += (diff * .5)
                    pos[1] += (.5 * lheight * (len(lines) - 1)) - (.5 * lheight)
                elif self.text_valign == 'bottom':
                    pos[1] += (lheight * (len(lines) -2))
                if self.parent and self.parent._draw_angle:
                    pos = rotate2d(pos, self.pivot, self.parent.angle)
                blf.position(font, pos[0], pos[1] - (charsize[1] * i * self.line_height), 0)
                blf.draw(font, txt)
        else:
            dimensions = blf.dimensions(font, self.text)
            pos = self._draw_pos.copy()
            # underground = dimensions[1] > charsize[1]

            if self.text_halign == 'center':
                pos[0] -= (dimensions[0] * .5)
            elif self.text_halign == 'right':
                pos[0] -= dimensions[0]
            if self.text_valign == 'top':
                # if underground:
                    # pos[1] += diff
                pos[1] -= charsize[1] * self.line_height
            elif self.text_valign == 'center':
                # if underground:
                    # pos[1] += (diff * .5)
                pos[1] -= (.5 * charsize[1])
            if self.parent and self.parent._draw_angle:
                pos = rotate2d(pos, self.pivot, self.parent.angle)
            blf.position(font, pos[0], pos[1], 0)
            blf.draw(font, self.text)

        super().draw()
        self.lines = lines
        blf.disable(font, blf.WORD_WRAP)
        blf.disable(font, blf.SHADOW)
        blf.disable(font, blf.ROTATION)



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULLimitRange(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.threshold = Vector((0, 0))
        self.min_value = None
        self.max_value = None
        self.operator = None
        self.last_val = 0
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        v = self.get_input(self.value)
        t = self.get_input(self.threshold)
        if self.min_value is not None:
            t = Vector((self.get_input(self.min_value), self.get_input(self.max_value)))
        else:
            t = self.get_input(self.threshold)
        self.calc_threshold(self.operator, v, t)
        return self.last_val

    def calc_threshold(self, op, v, t):
        last = self.last_val
        if op == 'OUTSIDE':
            if (v < t.x or v > t.y):
                self.last_val = v
            else:
                self.last_val = t.x if last <= t.x else t.y
        if op == 'INSIDE':
            if (t.x < v < t.y):
                self.last_val = v
            else:
                self.last_val = t.x if v <= t.x else t.y



from uplogic.nodes import ULParameterNode


class ULObjectDataName(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = self.add_output(self.get_name)

    def get_name(self):
        obj = self.get_input(self.game_object)
        return obj.blenderObject.name
